#include <iostream>
#include <iomanip>
#include <cpgplot.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <Utils/mycpgplot.hh>
#include <pvo-utils.hh>
using namespace mycpgplot;
using namespace PVO;
double angSepStd(double ra1, double dec1, double ra2, double dec2);
double angSepHav(double ra1, double dec1, double ra2, double dec2);
int main()
{
  
//   const int num=10000000;

  srandom(time(0));

  const int num=100000;
  float std[num],hav[num];
  float min,max;

  const float ramax = 1.;
  const float demax = 1.;

  double ra1  = dmsToDec("18:43:22.0")*15.;
  double dec1 = dmsToDec("89:52:12.3");
  double ra2  = dmsToDec("18:43:22.0")*15.;
  double dec2 = dmsToDec("89:52:12.1");
  
  std::cout << std::setprecision(10);
  std::cout << "Std result =\t" << angSepStd(ra1,dec1,ra2,dec2) << "\n";
  std::cout << "Hav result =\t" << angSepHav(ra1,dec1,ra2,dec2) << "\n";
  std::cout << "Should get  \t" << 0.2/3600. << "\n";

  for(int i=0;i<num;i++){
    
    ra1  = ramax * rand() / (RAND_MAX+1.);
    dec1 = demax * rand() / (RAND_MAX+1.) - demax/2.;
    ra2  = ramax * rand() / (RAND_MAX+1.);
    dec2 = demax * rand() / (RAND_MAX+1.) - demax/2.;

//     float std = angSepStd(ra1,dec1,ra2,dec2);
//     float hav = angSepHav(ra1,dec1,ra2,dec2);

    std[i] = angSepStd(ra1,dec1,ra2,dec2);
    hav[i] = angSepHav(ra1,dec1,ra2,dec2);

    if(i==0) min=max=std[i];
    else{
      min=std::min(min,std[i]);
      max=std::max(max,std[i]);
    }
    min=std::min(min,hav[i]);
    max=std::max(max,hav[i]);

    if(fabsf(std[i]-hav[i])>5.e-9){
      std::cout << "(" << ra1 << "," << dec1<< ") vs. (" << ra2 << "," << dec2 << ")"
		<< "  " << std[i] << " " << hav[i] << "\n";
    }

  }

  cpgopen("/vcps");
  cpgpap(8.,1.);
  cpgenv(min,max,min,max,0,0);
  cpglab("Standard way", "Haversine Formula","");
  cpgpt(num,std,hav,SOLIDCIRCLE);
  cpgmove(min,min);
  cpgdraw(max,max);

  float resid[num],rmin,rmax;
  for(int i=0;i<num;i++){
    resid[i] = std[i] - hav[i];
//     if(i==0) rmin=max=resid[i];
//     else{
//       rmin=std::min(rmin,resid[i]);
//       rmax=std::max(rmax,resid[i]);
//     }
  }
  rmin = -1.e-6;
  rmax = 1.e-6;
  cpgenv(min,max,rmin,rmax,0,0);
  cpglab("Standard way", "Standard - Haversine","");
  cpgpt(num,std,resid,SOLIDCIRCLE);
  cpgmove(min,0);
  cpgdraw(max,0);

  for(int i=0;i<num;i++){
    std[i] = log10(std[i]);
    hav[i] = log10(hav[i]);
    resid[i] = std[i] - hav[i];
//     if(i==0) rmin=max=resid[i];
//     else{
//       rmin=std::min(rmin,resid[i]);
//       rmax=std::max(rmax,resid[i]);
//     }
  }
  cpgenv(log10(min),log10(max),log10(min),log10(max),0,30);
  cpglab("log10(Standard way)", "log10(Haversine Formula)","");
  cpgpt(num,std,hav,SOLIDCIRCLE);
  cpgmove(log10(min),log10(min));
  cpgdraw(log10(max),log10(max));

  
  cpgenv(log10(min),log10(max),rmin,rmax,0,10);
  cpglab("log10(Standard way)", "log10(Standard/Haversine)","");
  cpgpt(num,std,resid,SOLIDCIRCLE);
  cpgmove(log10(min),0);
  cpgdraw(log10(max),0);

  cpgend();


}

double angSepStd(double ra1, double dec1, double ra2, double dec2)
{
    /**
     *  This is the "standard" way of calculating the angular separation.
     */
    double r1,d1,r2,d2;
    r1 = ra1  * M_PI / 180.;
    d1 = dec1 * M_PI / 180.;
    r2 = ra2  * M_PI / 180.;
    d2 = dec2 * M_PI / 180.;
//     std::cerr << cos(r1-r2) << " " << cos(d1) << " " << cos(d2) << " " << sin(d1) << " " << sin(d2) << "\n";
    double angsep = cos(r1-r2)*cos(d1)*cos(d2) + sin(d1)*sin(d2);
    
    return acos(angsep)*180./M_PI;
}

double angSepHav(double ra1, double dec1, double ra2, double dec2)
{
    /**
     *  This is the way of calculating the angular separation that uses the Haversine formula:
     *  see http://en.wikipedia.org/wiki/Haversine_formula
     *  and http://mathforum.org/library/drmath/view/51879.html for derivation.
     */
    double r1,d1,r2,d2,diffr,diffd;
    // put into radians
    r1 = ra1  * M_PI / 180.;
    d1 = dec1 * M_PI / 180.;
    r2 = ra2  * M_PI / 180.;
    d2 = dec2 * M_PI / 180.;
    diffr = (r1-r2)/2.;
    diffd = (d1-d2)/2.;
//     std::cerr << sin(diffr) << " " << cos(d1) << " " << cos(d2) << " " << sin(diffd) << "\n";
    double havtheta = sin(diffr)*sin(diffr)*cos(d1)*cos(d2) + sin(diffd)*sin(diffd);
      
    return asin(sqrt(havtheta))*2.*180./M_PI;
}
